/*
 *
 *    ******************************************************************************
 *    *
 *    * This program and the accompanying materials are made available under the
 *    * terms of the Apache License, Version 2.0 which is available at
 *    * https://www.apache.org/licenses/LICENSE-2.0.
 *    *
 *    *  See the NOTICE file distributed with this work for additional
 *    *  information regarding copyright ownership.
 *    * Unless required by applicable law or agreed to in writing, software
 *    * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
 *    * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
 *    * License for the specific language governing permissions and limitations
 *    * under the License.
 *    *
 *    * SPDX-License-Identifier: Apache-2.0
 *    *****************************************************************************
 *
 */

package net.brutex.ai.dnn.api;

import org.nd4j.linalg.api.ndarray.INDArray;
import org.nd4j.linalg.dataset.api.DataSet;
import org.nd4j.linalg.dataset.api.MultiDataSet;

/**
 * A Neural Network is an instance of a {@link INeuralNetworkConfiguration}, that can be trained,
 * evaluated, saved, exported, etc. Its configuration state is defined with the
 * {@link #setConfiguration(INeuralNetworkConfiguration)} and {@link #getConfiguration()} methods.
 *
 */
public interface IModel {

  /**
   * The configuration that defines this Neural Network
   *
   * @param conf the configuration to use for this network
   */
  void setConfiguration(INeuralNetworkConfiguration conf);
  INeuralNetworkConfiguration getConfiguration();

  /**
   * Fit the model for one iteration on the provided data
   *
   * @param features   the examples to classify (one example in each row)
   * @param labels the example labels(a binary outcome matrix)
   * @param featuresMask The mask array for the features (used for variable length time series, etc). May be null.
   * @param labelsMask The mask array for the labels (used for variable length time series, etc). May be null.
   */
  void fit(INDArray features, INDArray labels, INDArray featuresMask, INDArray labelsMask);

  /**
   * This method fits model with a given DataSet
   *
   * @param dataSet the dataset to use for training
   */
  void fit(DataSet dataSet);

  /**
   * This method fits model with a given MultiDataSet
   *
   * @param dataSet the multi dataset to use for training
   */
  void fit(MultiDataSet dataSet);

  /**
   * The name of the Neural Network
   * @return the name
   */
  String getName();

  /**
   * Set the name for this Neural Network
   * @param name the name
   */
  void setName(String name);

  /**
   * An implementation should provide a method to validate the network
   * @return true if no errors found; false otherwise
   */
  boolean isValid();

}
