/*
 *
 *    ******************************************************************************
 *    *
 *    * This program and the accompanying materials are made available under the
 *    * terms of the Apache License, Version 2.0 which is available at
 *    * https://www.apache.org/licenses/LICENSE-2.0.
 *    *
 *    *  See the NOTICE file distributed with this work for additional
 *    *  information regarding copyright ownership.
 *    * Unless required by applicable law or agreed to in writing, software
 *    * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
 *    * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
 *    * License for the specific language governing permissions and limitations
 *    * under the License.
 *    *
 *    * SPDX-License-Identifier: Apache-2.0
 *    *****************************************************************************
 *
 */

package net.brutex.ai.dnn.conf;

import com.fasterxml.jackson.annotation.JsonIgnoreProperties;
import com.fasterxml.jackson.databind.JsonNode;
import com.fasterxml.jackson.databind.ObjectMapper;
import com.fasterxml.jackson.databind.exc.InvalidTypeIdException;
import com.fasterxml.jackson.databind.node.ArrayNode;
import java.io.IOException;
import java.io.Serializable;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Map;
import java.util.Random;
import lombok.Getter;
import lombok.NonNull;
import lombok.Setter;
import lombok.Singular;
import lombok.extern.jackson.Jacksonized;
import lombok.extern.slf4j.Slf4j;
import net.brutex.ai.dnn.api.ILayerConfiguration;
import net.brutex.ai.dnn.api.INeuralNetworkConfiguration;
import org.deeplearning4j.nn.conf.BackpropType;
import org.deeplearning4j.nn.conf.CacheMode;
import org.deeplearning4j.nn.conf.GradientNormalization;
import org.deeplearning4j.nn.conf.InputPreProcessor;
import org.deeplearning4j.nn.conf.MultiLayerConfiguration;
import org.deeplearning4j.nn.conf.NeuralNetConfiguration;
import org.deeplearning4j.nn.conf.WorkspaceMode;
import org.deeplearning4j.nn.conf.distribution.Distribution;
import org.deeplearning4j.nn.conf.inputs.InputType;
import org.deeplearning4j.nn.conf.layers.BaseLayer;
import org.deeplearning4j.nn.conf.layers.BaseOutputLayer;
import org.deeplearning4j.nn.conf.layers.Layer;
import org.deeplearning4j.nn.conf.memory.LayerMemoryReport;
import org.deeplearning4j.nn.conf.memory.MemoryReport;
import org.deeplearning4j.nn.conf.memory.NetworkMemoryReport;
import org.deeplearning4j.nn.conf.serde.JsonMappers;
import org.deeplearning4j.nn.weights.IWeightInit;
import org.deeplearning4j.nn.weights.WeightInit;
import org.nd4j.linalg.activations.Activation;
import org.nd4j.linalg.activations.IActivation;
import org.nd4j.linalg.api.buffer.DataType;
import org.nd4j.linalg.learning.config.IUpdater;
import org.nd4j.linalg.lossfunctions.LossFunctions;
import org.nd4j.linalg.lossfunctions.impl.LossBinaryXENT;
import org.nd4j.linalg.lossfunctions.impl.LossMCXENT;
import org.nd4j.linalg.lossfunctions.impl.LossMSE;
import org.nd4j.linalg.lossfunctions.impl.LossNegativeLogLikelihood;

/**
 * The INeuralNetworkConfiguration is a sequential container for the different layers in your
 * network (or other NeuralNetworkConfigurations). That said, NeuralNetworkConfigurations can be
 * stacked.<br/><br/>
 * It then “chains” outputs to inputs sequentially for each INeuralNetworkConfiguration,
 * finally returning the output of the "top" configuration. Any settings made, are inherited and can
 * be overridden on a "deeper" level. For this use case, you need to wrap the INeuralNetworkConfiguration
 * into a BuildingBlockLayer
 *
 */
@Jacksonized
@JsonIgnoreProperties(ignoreUnknown = true)
@lombok.Builder
@Slf4j
public class NeuralNetworkConfiguration extends NeuralNetConfiguration implements
    INeuralNetworkConfiguration, Serializable, Cloneable {

    private static final int DEFAULT_TBPTT_LENGTH = 20;
  @Getter protected final List<NeuralNetworkConfiguration> confs = new ArrayList<>();
  /**
   * hidden list of layers, that "flattens" all the layers of this network and applies
   * inheritance.
   */
  @lombok.Builder.ObtainVia(method = "calculateInnerLayers")
  private final List<ILayerConfiguration> innerLayerConfigurations;
  @Getter @Setter @NonNull @Singular
  protected List<Layer> layers = new ArrayList<>();
  @Getter @Setter @NonNull @lombok.Builder.Default @Deprecated
  protected WorkspaceMode trainingWorkspaceMode = WorkspaceMode.ENABLED;
  @Getter @Setter @NonNull @lombok.Builder.Default @Deprecated
  protected WorkspaceMode inferenceWorkspaceMode = WorkspaceMode.ENABLED;
  /**
   * The type of backprop. Default setting is used for most networks (MLP, CNN etc), but
   * optionally truncated BPTT can be used for training recurrent neural networks. If using
   * TruncatedBPTT make sure you set both tBPTTForwardLength() and tBPTTBackwardLength()
   */
  @Getter @Setter @NonNull @lombok.Builder.Default
  protected BackpropType backpropType = BackpropType.Standard;
  @Getter
  protected Map<Integer, InputPreProcessor> inputPreProcessors = new HashMap<>();
  /**
   * When doing truncated BPTT: how many steps of forward pass should we do before doing
   * (truncated) backprop?<br> Only applicable when doing
   * backpropType(BackpropType.TruncatedBPTT)<br> Typically tBPTTForwardLength parameter is same
   * as the tBPTTBackwardLength parameter, but may be larger than it in some circumstances (but
   * never smaller)<br> Ideally your training data time series length should be divisible by this
   * This is the k1 parameter on pg23 of
   * <a
   * href="http://www.cs.utoronto.ca/~ilya/pubs/ilya_sutskever_phd_thesis.pdf">http://www.cs.utoronto.ca/~ilya/pubs/ilya_sutskever_phd_thesis.pdf</a>
   *
   * @param forwardLength Forward length > 0, >= backwardLength
   */
  @Getter @Setter protected int tbpttFwdLength = 20;
  /**
   * When doing truncated BPTT: how many steps of backward should we do?<br> Only applicable when
   * doing backpropType(BackpropType.TruncatedBPTT)<br> This is the k2 parameter on pg23 of
   * <a
   * href="http://www.cs.utoronto.ca/~ilya/pubs/ilya_sutskever_phd_thesis.pdf">http://www.cs.utoronto.ca/~ilya/pubs/ilya_sutskever_phd_thesis.pdf</a>
   *
   * @param backwardLength <= forwardLength
   */
  @Getter @Setter protected int tbpttBackLength = 20;
  /**
   * Creates and returns a copy of this object.
   *
   * @return a clone of this instance.
   * @throws CloneNotSupportedException if the object's class does not support the {@code Cloneable}
   *                                    interface. Subclasses that override the {@code clone} method
   *                                    can also throw this exception to indicate that an instance
   *                                    cannot be cloned.
   * @see Cloneable
   */

  //Nd4j.getRandom().setSeed(getConf(0).getSeed()); //TODO
  //Counter for the number of parameter updates so far
  // This is important for learning rate schedules, for example, and is stored here to ensure it is persisted
  // for Spark and model serialization
  @Getter @Setter
  protected int iterationCount = 0;
  //Counter for the number of epochs completed so far. Used for per-epoch schedules
  @Getter @Setter
  protected int epochCount = 0;
  protected double dampingFactor = 100;
  @Getter @Setter //todo why?
  private Layer layer;
  /**
   * A seed for this network, will be random if not specified.
    */
  @Getter @Setter @NonNull @lombok.Builder.Default
  private long seed = new Random().nextLong();
  /**
   * The default {@link CacheMode} for this configuration. Will be set to "NONE" if not specified otherwise.
   * This method defines how/if preOutput cache is handled: NONE: cache disabled (default value)
   * HOST: Host memory will be used DEVICE: GPU memory will be used (on CPU backends effect will
   * be the same as for HOST)
   *
   * Valid values are<br/>
   * CacheMode.NONE,<br/>
   * CacheMode.HOST or<br/>
   * CacheMode.DEVICE<br/>
   * @param cacheMode
   */
  @NonNull @Getter @Setter
  @lombok.Builder.Default private CacheMode cacheMode = CacheMode.NONE;
  /**
   * The list of layer configurations in this configuration. They will be indexed automatically
   * as the layers get added starting with index 0.
   */
  @Singular @Getter
  private List<ILayerConfiguration> layerConfigurations;
  /**
   * The name for this configuration. Defaults to "Anonymous INeuralNetworkConfiguration" if
   * it is not specified.
   */
  @lombok.Builder.Default @Getter
  private String name = "Anonymous INeuralNetworkConfiguration";
  /**
   * The {@link InputType} of the data for this network configuration
   */
  private InputType inputType;
  /**
   * Set the DataType for the network parameters and activations for all layers in the network.
   * Default: Float
   *
   * @param dataType Datatype to use for parameters and activations
   */
  @Getter @Setter @lombok.Builder.Default @NonNull
  private DataType dataType = DataType.FLOAT;
    /**
     * Whether to override the nIn configuration forcibly upon construction. Default value is true.
     * @return builder pattern
     */
    @Getter @Setter
    @lombok.Builder.Default
    private boolean overrideNinUponBuild = true;
    /**
     * Enabled by default. If enabled, the output layer configuration will be validated, to throw an
     * exception on likely invalid outputs - such as softmax + nOut=1, or LossMCXENT + Tanh.<br> If
     * disabled (false) no output layer validation will be performed.<br> Disabling this validation
     * is not recommended, as the configurations that fail validation usually will not be able to
     * learn correctly. However, the option to disable this validation is provided for advanced
     * users when creating non-standard architectures.
     *
     * @param validate If true: validate output layer configuration. False: don't validate
     */
    @Getter @Setter @lombok.Builder.Default
    private boolean validateOutputLayerConfig=true;
    /**
     * Enabled by default. If enabled, an exception will be throw when using the (invalid)
     * combination of truncated backpropagation through time (TBPTT) with either a
     * GlobalPoolingLayer or LastTimeStepLayer.<br> It is possible to disable this validation to
     * allow what is almost certainly an invalid configuration to be used, however this is not
     * recommended.
     *
     * @param validate Whether TBPTT validation should be performed
     */
    @Getter @Setter @lombok.Builder.Default
    private boolean validateTbpttConfig=true;



  /**
   * Gradient updater configuration. For example, {@link org.nd4j.linalg.learning.config.Adam}
   * or {@link org.nd4j.linalg.learning.config.Nesterovs}<br>
   * Note: values set by this method will be applied to all applicable layers in the network, unless a different
   * value is explicitly set on a given layer. In other words: values set via this method are used as the default
   * value, and can be overridden on a per-layer basis.
   *
   * @param updater Updater to use
   */
  @Getter @Setter @NonNull
  private IUpdater updater;

  /**
   * Gradient normalization strategy. Used to specify gradient renormalization, gradient clipping etc.
   * See {@link GradientNormalization} for details<br>
   * Note: values set by this method will be applied to all applicable layers in the network, unless a different
   * value is explicitly set on a given layer. In other words: values set via this method are used as the default
   * value, and can be overridden on a per-layer basis.
   *
   * @param gradientNormalization Type of normalization to use. Defaults to None.
   * @see GradientNormalization
   */
  @Getter @Setter @NonNull @lombok.Builder.Default
  private GradientNormalization gradientNormalization = GradientNormalization.None;

  /**
   * Threshold for gradient normalization, only used for GradientNormalization.ClipL2PerLayer,
   * GradientNormalization.ClipL2PerParamType, and GradientNormalization.ClipElementWiseAbsoluteValue<br>
   * Not used otherwise.<br>
   * L2 threshold for first two types of clipping, or absolute value threshold for last type of clipping.<br>
   * Note: values set by this method will be applied to all applicable layers in the network, unless a different
   * value is explicitly set on a given layer. In other words: values set via this method are used as the default
   * value, and can be overridden on a per-layer basis.
   */
  @Getter @Setter
  private double gradientNormalizationThreshold;


  /**
   * Weight initialization scheme to use, for initial weight values
   * Note: values set by this method will be applied to all applicable layers in the network, unless a different
   * value is explicitly set on a given layer. In other words: values set via this method are used as the default
   * value, and can be overridden on a per-layer basis.
   */
  @Getter @Setter
  private IWeightInit weightInit;

  /**
   * Activation function / neuron non-linearity<br>
   * Note: values set by this method will be applied to all applicable layers in the network, unless a different
   * value is explicitly set on a given layer. In other words: values set via this method are used as the default
   * value, and can be overridden on a per-layer basis.
   */
  @Getter @Setter
  private IActivation activation;



  /**
   * Create a neural net configuration from json
   *
   * @param json the neural net configuration from json
   * @return {@link NeuralNetworkConfiguration}
   */
  public static NeuralNetworkConfiguration fromJson(String json) {
    NeuralNetworkConfiguration conf;
    ObjectMapper mapper = NeuralNetworkConfiguration.mapper();
    try {
      conf = mapper.readValue(json, NeuralNetworkConfiguration.class);
    } catch (InvalidTypeIdException e) {
      if (e.getMessage().contains("@class")) {
        try {
          //JSON may be legacy (1.0.0-alpha or earlier), attempt to load it using old format
          return JsonMappers.getLegacyMapper().readValue(json, NeuralNetworkConfiguration.class);
        } catch (InvalidTypeIdException e2) {
          //Check for legacy custom layers: "Could not resolve type id 'CustomLayer' as a subtype of [simple type, class org.deeplearning4j.nn.conf.layers.ILayer]: known type ids = [Bidirectional, CenterLossOutputLayer, CnnLossLayer, ..."
          //1.0.0-beta5: dropping support for custom layers defined in pre-1.0.0-beta format. Built-in layers from these formats still work
          String msg = e2.getMessage();
          if (msg != null && msg.contains("Could not resolve type id")) {
            throw new RuntimeException(
                "Error deserializing MultiLayerConfiguration - configuration may have a custom " +
                    "layer, vertex or preprocessor, in pre version 1.0.0-beta JSON format.\nModels in legacy format with custom"
                    +
                    " layers should be loaded in 1.0.0-beta to 1.0.0-beta4 and saved again, before loading in the current version of DL4J",
                e);
          }
          throw new RuntimeException(e2);
        } catch (IOException e2) {
          throw new RuntimeException(e2);
        }
      }
      throw new RuntimeException(e);
    } catch (IOException e) {
      //Check if this exception came from legacy deserializer...
      String msg = e.getMessage();
      if (msg != null && msg.contains("legacy")) {
        throw new RuntimeException(
            "Error deserializing MultiLayerConfiguration - configuration may have a custom " +
                "layer, vertex or preprocessor, in pre version 1.0.0-alpha JSON format. These layers can be "
                +
                "deserialized by first registering them with NeuralNetConfiguration.registerLegacyCustomClassesForJSON(Class...)",
            e);
      }
      throw new RuntimeException(e);
    }

    //To maintain backward compatibility after loss function refactoring (configs generated with v0.5.0 or earlier)
    // Previously: enumeration used for loss functions. Now: use classes
    // IN the past, could have only been an OutputLayer or RnnOutputLayer using these enums
    int layerCount = 0;
    JsonNode confs = null;
    for (NeuralNetworkConfiguration nnc : conf.getConfs()) {
      Layer l = nnc.getLayer();
      if (l instanceof BaseOutputLayer && ((BaseOutputLayer) l).getLossFn() == null) {
        //lossFn field null -> may be an old config format, with lossFunction field being for the enum
        //if so, try walking the JSON graph to extract out the appropriate enum value

        BaseOutputLayer ol = (BaseOutputLayer) l;
        try {
          JsonNode jsonNode = mapper.readTree(json);
          if (confs == null) {
            confs = jsonNode.get("confs");
          }
          if (confs instanceof ArrayNode) {
            ArrayNode layerConfs = (ArrayNode) confs;
            JsonNode outputLayerNNCNode = layerConfs.get(layerCount);
            if (outputLayerNNCNode == null) {
              throw new RuntimeException("should never happen"); //return conf; //Should never happen...
            }
            JsonNode outputLayerNode = outputLayerNNCNode.get("layer");

            JsonNode lossFunctionNode = null;
            if (outputLayerNode.has("output")) {
              lossFunctionNode = outputLayerNode.get("output").get("lossFunction");
            } else if (outputLayerNode.has("rnnoutput")) {
              lossFunctionNode = outputLayerNode.get("rnnoutput").get("lossFunction");
            }

            if (lossFunctionNode != null) {
              String lossFunctionEnumStr = lossFunctionNode.asText();
              LossFunctions.LossFunction lossFunction = null;
              try {
                lossFunction = LossFunctions.LossFunction.valueOf(lossFunctionEnumStr);
              } catch (Exception e) {
                log.warn(
                    "OutputLayer with null LossFunction or pre-0.6.0 loss function configuration detected: could not parse JSON",
                    e);
              }

              if (lossFunction != null) {
                switch (lossFunction) {
                  case MSE:
                    ol.setLossFn(new LossMSE());
                    break;
                  case XENT:
                    ol.setLossFn(new LossBinaryXENT());
                    break;
                  case NEGATIVELOGLIKELIHOOD:
                    ol.setLossFn(new LossNegativeLogLikelihood());
                    break;
                  case MCXENT:
                    ol.setLossFn(new LossMCXENT());
                    break;

                  //Remaining: TODO
                  case SQUARED_LOSS:
                  case RECONSTRUCTION_CROSSENTROPY:
                  default:
                    log.warn(
                        "OutputLayer with null LossFunction or pre-0.6.0 loss function configuration detected: could not set loss function for {}",
                        lossFunction);
                    break;
                }
              }
            }

          } else {
            log.warn(
                "OutputLayer with null LossFunction or pre-0.6.0 loss function configuration detected: could not parse JSON: layer 'confs' field is not an ArrayNode (is: {})",
                (confs != null ? confs.getClass() : null));
          }
        } catch (IOException e) {
          log.warn(
              "OutputLayer with null LossFunction or pre-0.6.0 loss function configuration detected: could not parse JSON",
              e);
          break;
        }
      }

      //Also, pre 0.7.2: activation functions were Strings ("activationFunction" field), not classes ("activationFn")
      //Try to load the old format if necessary, and create the appropriate IActivation instance
      if ((l instanceof BaseLayer) && ((BaseLayer) l).getActivationFn() == null) {
        try {
          JsonNode jsonNode = mapper.readTree(json);
          if (confs == null) {
            confs = jsonNode.get("confs");
          }
          if (confs instanceof ArrayNode) {
            ArrayNode layerConfs = (ArrayNode) confs;
            JsonNode outputLayerNNCNode = layerConfs.get(layerCount);
            if (outputLayerNNCNode == null) {
              throw new RuntimeException("Should never happen"); //return conf; //Should never happen...
            }
            JsonNode layerWrapperNode = outputLayerNNCNode.get("layer");

            if (layerWrapperNode == null || layerWrapperNode.size() != 1) {
              continue;
            }

            JsonNode layerNode = layerWrapperNode.elements().next();
            JsonNode activationFunction = layerNode.get(
                "activationFunction"); //Should only have 1 element: "dense", "output", etc

            if (activationFunction != null) {
              IActivation ia = Activation.fromString(activationFunction.asText())
                  .getActivationFunction();
              ((BaseLayer) l).setActivationFn(ia);
            }
          }

        } catch (IOException e) {
          log.warn(
              "ILayer with null ActivationFn field or pre-0.7.2 activation function detected: could not parse JSON",
              e);
        }
      }

      if (!handleLegacyWeightInitFromJson(json, l, mapper, confs, layerCount)) {
        return conf;
      }

      layerCount++;
    }
    return conf;
  }

  /**
   * Handle {@link WeightInit} and {@link Distribution} from legacy configs in Json format. Copied
   * from handling of {@link Activation} above.
   *
   * @return True if all is well and layer iteration shall continue. False else-wise.
   */
  private static boolean handleLegacyWeightInitFromJson(String json, Layer l, ObjectMapper mapper,
      JsonNode confs, int layerCount) {
    if ((l instanceof BaseLayer) && ((BaseLayer) l).getWeightInitFn() == null) {
      try {
        JsonNode jsonNode = mapper.readTree(json);
        if (confs == null) {
          confs = jsonNode.get("confs");
        }
        if (confs instanceof ArrayNode) {
          ArrayNode layerConfs = (ArrayNode) confs;
          JsonNode outputLayerNNCNode = layerConfs.get(layerCount);
          if (outputLayerNNCNode == null) {
            return false; //Should never happen...
          }
          JsonNode layerWrapperNode = outputLayerNNCNode.get("layer");

          if (layerWrapperNode == null || layerWrapperNode.size() != 1) {
            return true;
          }

          JsonNode layerNode = layerWrapperNode.elements().next();
          JsonNode weightInit = layerNode.get(
              "weightInit"); //Should only have 1 element: "dense", "output", etc
          JsonNode distribution = layerNode.get("dist");

          Distribution dist = null;
          if (distribution != null) {
            dist = mapper.treeToValue(distribution, Distribution.class);
          }

          if (weightInit != null) {
            final IWeightInit wi = WeightInit.valueOf(weightInit.asText())
                .getWeightInitFunction(dist);
            ((BaseLayer) l).setWeightInitFn(wi);
          }
        }

      } catch (IOException e) {
        log.warn(
            "ILayer with null WeightInit detected: " + l.getLayerName() + ", could not parse JSON",
            e);
      }
    }
    return true;

  }

  /**
   * Object mapper for serialization of configurations
   *
   * @return
   */
  public static ObjectMapper mapperYaml() {
    return JsonMappers.getMapperYaml();
  }

  /**
   * Object mapper for serialization of configurations
   *
   * @return
   */
  public static ObjectMapper mapper() {
    return JsonMappers.getMapper();
  }



  /**
   * @return JSON representation of NN configuration
   */
  public String toYaml() {
    ObjectMapper mapper = NeuralNetConfiguration.mapperYaml();
    synchronized (mapper) {
      try {
        return mapper.writeValueAsString(this);
      } catch (com.fasterxml.jackson.core.JsonProcessingException e) {
        throw new RuntimeException(e);
      }
    }
  }

  /**
   * @return JSON representation of NN configuration
   */
  public String toJson() {
    ObjectMapper mapper = NeuralNetConfiguration.mapper();
    synchronized (mapper) {
      //JSON mappers are supposed to be thread safe: however, in practice they seem to miss fields occasionally
      //when writeValueAsString is used by multiple threads. This results in invalid JSON. See issue #3243
      try {
        return mapper.writeValueAsString(this);
      } catch (com.fasterxml.jackson.core.JsonProcessingException e) {
        throw new RuntimeException(e);
      }
    }
  }

  @Override
  public String toString() {
    return toJson();
  }

  public NeuralNetworkConfiguration getConf(int i) {
    return confs.get(i);
  }

  @Override
  public NeuralNetworkConfiguration clone() {

      NeuralNetworkConfiguration clone = (NeuralNetworkConfiguration) super.clone();
      List<NeuralNetworkConfiguration> confList = clone.getConfs();
      if (confList != null) {
        List<NeuralNetworkConfiguration> list = new ArrayList<>();
        for (NeuralNetworkConfiguration conf : confList) {
          list.add(conf.clone());
        }
      }

      if (clone.getInputPreProcessors() != null) {
        Map<Integer, InputPreProcessor> map = new HashMap<>();
        for (Map.Entry<Integer, InputPreProcessor> entry : clone.getInputPreProcessors().entrySet()) {
          map.put(entry.getKey(), entry.getValue().clone());
        }
        clone.getInputPreProcessors().clear();
        clone.getInputPreProcessors().putAll(map);
      }

      clone.setInferenceWorkspaceMode(this.inferenceWorkspaceMode);
      clone.setTrainingWorkspaceMode(this.trainingWorkspaceMode);
      clone.setCacheMode(this.cacheMode);
      clone.setValidateOutputLayerConfig(this.validateOutputLayerConfig);
      clone.setDataType(this.dataType);

      return clone;

  }

  public InputPreProcessor getInputPreProcess(int curr) {
    return inputPreProcessors.get(curr);
  }

  /**
   * Get a {@link MemoryReport} for the given MultiLayerConfiguration. This is used to estimate the
   * memory requirements for the given network configuration and input
   *
   * @param inputType Input types for the network
   * @return Memory report for the network
   */
  public NetworkMemoryReport getMemoryReport(InputType inputType) {

    Map<String, MemoryReport> memoryReportMap = new LinkedHashMap<>();
    int nLayers = confs.size();
    for (int i = 0; i < nLayers; i++) {
      String layerName = confs.get(i).getLayer().getLayerName();
      if (layerName == null) {
        layerName = String.valueOf(i);
      }

      //Pass input type through preprocessor, if necessary
      InputPreProcessor preproc = getInputPreProcess(i);
      //TODO memory requirements for preprocessor
      if (preproc != null) {
        inputType = preproc.getOutputType(inputType);
      }

      LayerMemoryReport report = confs.get(i).getLayer().getMemoryReport(inputType);
      memoryReportMap.put(layerName, report);

      inputType = confs.get(i).getLayer().getOutputType(i, inputType);
    }

    return new NetworkMemoryReport(memoryReportMap, MultiLayerConfiguration.class,
        "MultiLayerNetwork", inputType);
  }

  /**
   * For the given input shape/type for the network, return a list of activation sizes for each
   * layer in the network.<br> i.e., list.get(i) is the output activation sizes for layer i
   *
   * @param inputType Input type for the network
   * @return A lits of activation types for the network, indexed by layer number
   */
  public List<InputType> getLayerActivationTypes(@NonNull InputType inputType) {
    List<InputType> out = new ArrayList<>();
    int nLayers = confs.size();
    for (int i = 0; i < nLayers; i++) {
      InputPreProcessor preproc = getInputPreProcess(i);
      if (preproc != null) {
        inputType = preproc.getOutputType(inputType);
      }

      inputType = confs.get(i).getLayer().getOutputType(i, inputType);
      out.add(inputType);
    }
    return out;
  }

  /**
   * Defines some additional handy methods. Other than that,
   * the builder is generated by lombok.
   */
  public static class NeuralNetworkConfigurationBuilder {

      /**
       * Specify the processors. These are used at each layer for doing things like normalization and
       * shaping of input.
       *
       * @param processor what to use to preProcess the data.
       * @return builder pattern
       */
      public NeuralNetworkConfigurationBuilder inputPreProcessor(Integer layer,
          InputPreProcessor processor) {
        inputPreProcessors.put(layer, processor);
        return this;
      }

    /**
     * Specify additional layer configurations
     */
    @Deprecated
    public NeuralNetworkConfigurationBuilder layersFromArray(Layer[] arrLayers) {
      for(Layer l : arrLayers) {
        layers.add( l );
      }
      return this;
    }
  }


}
